def seasonal():
    return 1



"""

#https://towardsdatascience.com/time-series-analysis-creating-synthetic-datasets-cf008208e014
def seasonal(range):
    dates = list(pd.data_range(range))
    length = range.length()
    data = np.where(time < 10, time**3, (time-9)**2)
    
    for i in range 5:
        for j in range 50:
            seasonal.append(values[j])

"""